// *************************************************************************
//
// Copyright (C) 2004,2005 Bruno PAGES  All rights reserved.
//
// This file is part of the BOUML Uml Toolkit.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// e-mail : bouml@free.fr
// home   : http://bouml.free.fr
//
// *************************************************************************

#include <stdio.h>
#include <qtextstream.h> 

#include "UmlOperation.h"
#include "UmlSettings.h"
#include "JavaSettings.h"
#include "UmlCom.h"
//#include "JavaRefType.h"
#include "UmlClass.h"
#include "util.h"

/*
void UmlOperation::compute_dependency(QList<JavaRefType> & dependencies,
				      QCString cl_stereotype) {
  if (cl_stereotype == "enum")
    return;
  
  QCString decl = javaDecl();
  
  remove_comments(decl);
  remove_arrays(decl);
  
  int index;
  
  if ((index = decl.find("${visibility}")) != -1)
    decl.remove((unsigned) index, 13);
  if ((index = decl.find("${static}")) != -1)
    decl.remove((unsigned) index, 9);
  if ((index = decl.find("${final}")) != -1)
    decl.remove((unsigned) index, 8);
  if ((index = decl.find("${abstract}")) != -1)
    decl.remove((unsigned) index, 11);
  if ((index = decl.find("${synchronized}")) != -1)
    decl.remove((unsigned) index, 15);
  if ((index = decl.find("${name}")) != -1)
    decl.remove((unsigned) index, 7);
  if ((index = decl.find("${stereotype}")) != -1)
    decl.replace((unsigned) index, 13,
		 JavaSettings::relationStereotype(stereotype()));

  int index2;
  
  if (((index = decl.find("${(}")) == -1) ||
      ((index2 = decl.find("${)}", index)) == -1))
    return;
  
  decl.replace(index2, 4, ")");
  decl.replace(index, 4, "(");
  
  const char * p = decl;
  const char * dontsubstituteuntil = 0;
  const QValueList<UmlParameter> & params = this->params();
  UmlTypeSpec ts;
  
  for (;;) {
    char c;
    bool dontsearchend = FALSE;
    
    // search word beginning
    while ((c = *p) != 0) {
      if ((c == '_') ||
	  ((c >= 'a') && (c <= 'z')) ||
	  ((c >= 'A') && (c <= 'Z')))
	break;
      else if (dontsubstituteuntil != 0) {
	if (p >= dontsubstituteuntil)
	  dontsubstituteuntil = 0;
	p += 1;
      }
      else if (!strncmp(p, "${type}", 7)) {
	p += 7;
	ts = returnType();
	if (ts.type != 0) {
	  dontsearchend = TRUE;
	  break;
	}
	else {
	  decl = ts.explicit_type + p;
	  p = decl;
	}
      }
      else if (sscanf(p, "${t%u}", (unsigned *) &index) == 1) {
	p = strchr(p, '}') + 1;
	if (((unsigned) index) < params.count()) {
	  // else error signaled later
	  ts = params[index].type;
	  if (ts.type != 0) {
	    dontsearchend = TRUE;
	    break;
	  }
	  else {
	    decl = ts.explicit_type + p;
	    p = decl;
	  }
	}
      }
      else if (sscanf(p, "${p%u}", &index) == 1)
	p = strchr(p, '}') + 1;
      else if (c == '{')
	// body
	return;
      else
	p += 1;
    }
    
    if (c == 0)
      return;
    
    if (!dontsearchend) {
      // search word end
      const char * p2 = p;
      
      ts.explicit_type = p2;
      p += 1;
      
      while ((c = *p) != 0) {
	if ((c == '_') ||
	    (c == '.') ||
	    ((c >= 'a') && (c <= 'z')) ||
	    ((c >= 'A') && (c <= 'Z')) ||
	    ((c >= '0') && (c <= '9')))
	  p += 1;
	else {
	  ts.explicit_type.truncate(p - p2);
	  break;
	}
      }
    
      if ((p2 = strrchr(ts.explicit_type, '.')) != 0)
	// remove package name !!!
	ts.explicit_type = p2 + 1;
      
      if (dontsubstituteuntil == 0) {
	QCString subst = JavaSettings::type(ts.explicit_type);
	
	if (subst != ts.explicit_type) {
	  decl = subst + ' ' + p;
	  p = decl;
	  dontsubstituteuntil = p + subst.length();
	  continue;
	}
      }
    }
    
    // check manually added keyword
    if ((ts.explicit_type == "final") ||
	(ts.explicit_type == "static"))
      continue;
    
    JavaRefType::add(ts, dependencies);
  }
}
*/

static bool generate_type(const QValueList<UmlParameter> & params,
			  unsigned rank, QTextOStream & f)
{
  if (rank >= params.count())
    return FALSE;
  
  UmlClass::write(f, params[rank].type);
  return TRUE;
}

static bool generate_var(const QValueList<UmlParameter> & params, 
			 unsigned rank, QTextOStream & f)
{
  if (rank >= params.count())
    return FALSE;
  
  f << params[rank].name;
  return TRUE;
}

static void param_error(const QCString & parent, const QCString & name, unsigned rank)
{
  UmlCom::trace(QCString("<tt>        </tt><font face=helvetica><b>while compiling <i>")
		+ parent + "::" + name + "</i> parameter rank " + QCString().setNum(rank)
		+ " does not exist</font></b><br>");
}

QCString UmlOperation::compute_name() {
  QCString get_set_spec = javaNameSpec();
  
  if (! get_set_spec.isEmpty()) {
    UmlClassMember * it;
    
    if ((it = getOf()) == 0)
      it = setOf();
    
    int index;
    QCString s = it->name();
    
    if ((index = s.find(' ')) != -1)
      // remove " (relation_name)"
      s = s.left(index);
    
    if ((index = get_set_spec.find("${name}")) != -1)
      get_set_spec.replace(index, 7, s);
    else if ((index = get_set_spec.find("${Name}")) != -1)
      get_set_spec.replace(index, 7, capitalize(s));
    return get_set_spec;
  }
  else
    return name();
}

void UmlOperation::generate(QTextOStream & f, const QCString & cl_stereotype,
			    QCString indent) {
  if (!javaDecl().isEmpty()) {
    const char * p = javaDecl();
    const QValueList<UmlParameter> & params = this->params();
    unsigned rank;
    
    while ((*p == ' ') || (*p == '\t'))
      indent += *p++;
  
    f << indent;
    
    while (*p) {
      if (*p == '\n') {
	f << *p++;
	if (*p)
	  f << indent;
      }
      else if ((*p == '{') && 
	       (isAbstract() || (cl_stereotype == "interface"))) {
	f << ";";
	break;
      }
      else if (*p != '$')
	f << *p++;
      else if (!strncmp(p, "${comment}", 10))
	generate_comment(p, f, indent);
      else if (!strncmp(p, "${visibility}", 13)) {
	p += 13;
	generate_visibility(f);
      }
      else if (!strncmp(p, "${final}", 8)) {
	p += 8;
	if (isJavaFinal())
	  f << "final ";
      }
      else if (!strncmp(p, "${static}", 9)) {
	p += 9;
	if (isClassMember())
	  f << "static ";
      }
      else if (!strncmp(p, "${abstract}", 11)) {
	p += 11;
	if (isAbstract() && (cl_stereotype != "interface"))
	  f << "abstract ";
      }
      else if (!strncmp(p, "${synchronized}", 15)) {
	p += 15;
	if (isJavaSynchronized())
	  f << "synchronized ";
      }
      else if (!strncmp(p, "${type}", 7)) {
	p += 7;
	UmlClass::write(f, returnType());
      }
      else if (!strncmp(p, "${stereotype}", 13)) {
	p += 13;
	// get/set relation with multiplicity > 1
	f << JavaSettings::relationStereotype(stereotype());
      }
      else if (!strncmp(p, "${name}", 7)) {
	p += 7;
	f << compute_name();
      }
      else if (!strncmp(p, "${(}", 4)) {
	p += 4;
	f << '(';
      }
      else if (!strncmp(p, "${)}", 4)) {
	p += 4;
	f << ')';
      }
      else if (sscanf(p, "${t%u}", &rank) == 1) {
	if (!generate_type(params, rank, f))
	  param_error(parent()->name(), name(), rank);
	p = strchr(p, '}') + 1;
      }
      else if (sscanf(p, "${p%u}", &rank) == 1) {
	if (!generate_var(params, rank, f))
	  param_error(parent()->name(), name(), rank);
	p = strchr(p, '}') + 1;
      }
      else if (!strncmp(p, "${throws}", 9)) {
	p += 9;
	const char * sep;
	const QValueList<UmlTypeSpec> & exceptions = this->exceptions();
	QValueList<UmlTypeSpec>::ConstIterator it;
	
	for (it = exceptions.begin(), sep = " throws "; 
	     it != exceptions.end();
	     ++it, sep = " ") {
	  f << sep;
	  UmlClass::write(f, *it);
	}
      }
      else if (!strncmp(p, "${staticnl}", 11)) {
	p += 11;
	if (isClassMember()) {
	  f << '\n';
	  if (*p)
	    f << indent;
	}
	else
	  f << ' ';
      }
      else if (!strncmp(p, "${body}", 7)) {
	p += 7;
	
	QCString b = javaBody();
	
	if (!b.isEmpty()) {
	  for (const char * pb = b; *pb; pb += 1) {
	    f << *pb;
	    if (*pb == '\n')
	      f << indent;
	  }
	}
      }
      else
	// strange
	f << *p++;
    }
    
    f << '\n';
  }
}

void UmlOperation::generate_enum_item(QTextOStream &, int &,
				      const QCString &, QCString) {
  if (!javaDecl().isEmpty())
    UmlCom::trace("<tt>        </tt><font face=helvetica><b>an <i>enum</i> cannot have operation</b></font><br>");
}

void UmlOperation::generate_enum_case(QTextOStream &, QCString) {
  // error already signaled
}
