// *************************************************************************
//
// Copyright (C) 2004,2005 Bruno PAGES  All rights reserved.
//
// This file is part of the BOUML Uml Toolkit.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// e-mail : bouml@free.fr
// home   : http://bouml.free.fr
//
// *************************************************************************

#include <qtextstream.h> 

#include "UmlRelation.h"
#include "UmlClass.h"
#include "UmlSettings.h"
#include "JavaSettings.h"
//#include "JavaRefType.h"
#include "UmlCom.h"

/*
void UmlRelation::compute_dependency(QList<JavaRefType> & dependencies,
				     QCString cl_stereotype) {
  if (cl_stereotype == "enum")
    return;
  
  switch (relationKind()) {
  case aDependency:
  case aGeneralisation:
  case aRealization:
    JavaRefType::add(roleType(), dependencies);
    break;
  default:
    QCString decl = javaDecl();
    int index;
    
    remove_comments(decl);

    if ((index = decl.find("${visibility}")) != -1)
      decl.remove((unsigned) index, 13);
    if ((index = decl.find("${static}")) != -1)
      decl.remove((unsigned) index, 9);
    if ((index = decl.find("${final}")) != -1)
      decl.remove((unsigned) index, 8);
    if ((index = decl.find("${transient}")) != -1)
      decl.remove((unsigned) index, 12);
    if ((index = decl.find("${volatile}")) != -1)
      decl.remove((unsigned) index, 11);
    if ((index = decl.find("${name}")) != -1)
      decl.remove((unsigned) index, 7);
    if ((index = decl.find("${value}")) != -1)
      decl.remove((unsigned) index, 8);
    
    if ((index = decl.find("${stereotype}")) != -1)
      decl.replace((unsigned) index, 13,
		   JavaSettings::relationStereotype(stereotype()));
    
    UmlTypeSpec type;
    
    type.type = roleType();
    UmlClassMember::compute_dependency(dependencies, decl, type);
  }
}
*/

void UmlRelation::generate_extends(const char *& sep, QTextOStream & f,
				   const QCString & cl_stereotype) {
  switch (relationKind()) {
  default:
    return;
  case aGeneralisation:
  case aRealization:
    if (javaDecl().isEmpty())
      return;
    
    UmlClass * role_type = roleType();
    const QCString & other_stereotype = role_type->java_stereotype();
    
    if (cl_stereotype == "interface") {
      if (other_stereotype != "interface")
	UmlCom::trace(QCString("<tt>        </tt><font face=helvetica><b>cannot extends a <i>")
		      + other_stereotype + "</i></b></font><br>");
      else {
	f << sep;
	role_type->write(f);
	sep = ", ";
      }
    }
    else if (other_stereotype != "interface") {
      if ((cl_stereotype == "union") || (cl_stereotype == "enum"))
	UmlCom::trace(QCString("<tt>        </tt><font face=helvetica><b>an <i>")
		      + cl_stereotype + "</i> cannot extends</b></font><br>");
      else if ((other_stereotype == "union") || (other_stereotype == "enum"))
	UmlCom::trace(QCString("<tt>        </tt><font face=helvetica><b>cannot extends an <i>")
		      + other_stereotype + "</i></b></font><br>");
      else if (*sep == ',') {
	UmlCom::trace("<tt>        </tt><font face=helvetica><b>extend several classes</b></font><br>");
	return;
      }
      else {
	f << sep;
	role_type->write(f);
	sep = ", ";
      }
    }
  }
}

void UmlRelation::generate_implements(const char *& sep, QTextOStream & f,
				      const QCString & cl_stereotype) {
  switch (relationKind()) {
  default:
    return;
  case aGeneralisation:
  case aRealization:
    if (javaDecl().isEmpty()) 
      return;
    
    UmlClass * role_type = roleType();
    const QCString & other_stereotype = role_type->java_stereotype();
    
    if (other_stereotype == "interface") {
      if ((cl_stereotype == "union") || (cl_stereotype == "enum"))
	UmlCom::trace(QCString("<tt>        </tt><font face=helvetica><b>an <i>")
		      + cl_stereotype + "</i> cannot inherits</b></font><br>");
      else {
	f << sep;
	role_type->write(f);
	sep = ", ";
      }
    }
  }
}

void UmlRelation::generate(QTextOStream & f, const QCString & cl_stereotype,
			   QCString indent) {
  switch (relationKind()) {
  case aDependency:
  case aGeneralisation:
  case aRealization:
    break;
  default:
    if (!javaDecl().isEmpty()) {
      if (cl_stereotype == "enum") {
	UmlCom::trace("<tt>        </tt><font face=helvetica><b>an <i>enum</i> cannot have relation</b></font><br>");
	return;
      }
      
      const char * p = javaDecl();
      QCString s;
      
      while ((*p == ' ') || (*p == '\t'))
	indent += *p++;
      
      f << indent;
      
      while (*p) {
	if (*p == '\n') {
	  f << *p++;
	  if (*p)
	    f << indent;
	}
	else if (*p != '$')
	  f << *p++;
	else if (!strncmp(p, "${comment}", 10))
	  generate_comment(p, f, indent);
	else if (!strncmp(p, "${visibility}", 13)) {
	  p += 13;
	  generate_visibility(f);
	}
	else if (!strncmp(p, "${static}", 9)) {
	  p += 9;
	  if (isClassMember())
	    f << "static ";
	}
	else if (!strncmp(p, "${final}", 8)) {
	  p += 8;
	  if (isReadOnly())
	    f << "final ";
	}
	else if (!strncmp(p, "${transient}", 12)) {
	  p += 12;
	  if (isJavaTransient())
	    f << "transient ";
	}
	else if (!strncmp(p, "${volatile}", 11)) {
	  p += 11;
	  if (isVolatile())
	    f << "volatile ";
	}
	else if (!strncmp(p, "${type}", 7)) {
	  p += 7;
	  roleType()->write(f);
	}
	else if (!strncmp(p, "${stereotype}", 13)) {
	  p += 13;
	  f << JavaSettings::relationStereotype(stereotype());
	}
	else if (!strncmp(p, "${name}", 7)) {
	  p += 7;
	  f << roleName();
	}
	else if (!strncmp(p, "${multiplicity}", 15)) {
	  p += 15;
	  
	  QCString m = multiplicity();
	  
	  if (*m != '[')
	    f << "[]";
	  else {
	    for (unsigned index = 0; index != m.length(); index += 1) {
	      switch (m.at(index)) {
	      case '[':
		f << '[';
		break;
	      case ']':
		f << ']';
	      default:
		break;
	      }
	    }
	  }
	}
	else if (!strncmp(p, "${value}", 8)) {
	  p += 8;
	  if (!defaultValue().isEmpty()) {
	    if (*((const char *) defaultValue()) == '=')
	      f << ' ' << defaultValue();
	    else
	      f << defaultValue();
	  }
	}
	else
	  // strange
	  f << *p++;
      }
    
      f << '\n';
    }
  }
}

void UmlRelation::generate_enum_item(QTextOStream &, int &,
				      const QCString &, QCString) {
  if (!javaDecl().isEmpty())
    UmlCom::trace("<tt>        </tt><font face=helvetica><b>an <i>enum</i> cannot have relation</b></font><br>");
}

void UmlRelation::generate_enum_case(QTextOStream &, QCString) {
  // error already signaled
}


