// *************************************************************************
//
// Copyright (C) 2004,2005 Bruno PAGES  All rights reserved.
//
// This file is part of the BOUML Uml Toolkit.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// e-mail : bouml@free.fr
// home   : http://bouml.free.fr
//
// *************************************************************************

#ifndef UMLWINDOW_H
#define UMLWINDOW_H

#include <qmainwindow.h>
#include <qstringlist.h>

class QCanvasView;
class QWorkspace;
class QMultiLineEdit;
class QToolBar;
class QPopupMenu;

class BrowserView;
class BrowserNode;

// singleton

class UmlWindow : public QMainWindow {
  Q_OBJECT
    
  protected:
    static UmlWindow * the;
  
    QWorkspace * ws;
    BrowserView * browser;
    QMultiLineEdit * comment;
    BrowserNode * commented;	// the commented object
    QToolBar * projectTools;
    QPopupMenu * projectMenu;
    QPopupMenu * windowsMenu;
    QPopupMenu * toolMenu;
    QPopupMenu * miscMenu;
    QPopupMenu * fontSizeMenu;
    int use_cpp_id;
    int use_java_id;
    int use_idl_id;
    int show_browser_stereotypes_id;
    QStringList historic;

  public:
    UmlWindow();
    ~UmlWindow();
  
    void load(QString fn);
    bool can_close();
    void close_it();
    static void load_it(QString fn);
    
    static void set_commented(BrowserNode * bn);
    static void set_message(const QString &);
    static QWorkspace * get_workspace();
    static void clear();
    static void historic_add(QString fn);
    static void save_it();

  private slots:
    void newProject();
    void load();
    void save();
    void saveAs();
    void print();
    void close();
    void quit();
    void browser_search();
    
    void use_cpp();
    void use_java();
    void use_idl();
    
    void motif_style();
    void motifplus_style();
    //void sgi_style();
    //void cde_style();
    void windows_style();
    
    void show_stereotypes();

    void about();
    void aboutQt();
    
    void show_trace();
    void cpp_generate();
    void java_generate();
    void idl_generate();
    void java_catalog();
    //void doc_generate();
    void cpp_reverse();
    void java_reverse();
    void run_tool(int param);
    void tool_settings();
    void import_tool_settings();
    
    void comment_changed();
    void preferred_geometry();

    void toolMenuAboutToShow();
    void projectMenuAboutToShow();
    void historicActivated(int id);
    
    void miscMenuAboutToShow();
    
    void fontSizeMenuAboutToShow();
    void setFontSize(int);

    void windowsMenuAboutToShow();
    void windowsMenuActivated(int id);
};

#endif
