// *************************************************************************
//
// Copyright (C) 2004,2005 Bruno PAGES  All rights reserved.
//
// This file is part of the BOUML Uml Toolkit.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// e-mail : bouml@free.fr
// home   : http://bouml.free.fr
//
// *************************************************************************

#include <stdio.h>

#include <qcursor.h>
#include <qgrid.h>
#include <qvbox.h>
#include <qlabel.h>
#include <qcombobox.h> 
#include <qbuttongroup.h>
#include <qcheckbox.h>
#include <qradiobutton.h> 
#include <qpopupmenu.h> 
#include <qpushbutton.h> 
#include <qmessagebox.h>

#include "OperationDialog.h"
#include "OperationData.h"
#include "ParamData.h"
#include "BrowserOperation.h"
#include "ClassData.h"
#include "BrowserView.h"
#include "BrowserClass.h"
#include "KeyValueTable.h"
#include "ComboItem.h"
#include "myio.h"
#include "UmlWindow.h"
#include "DialogUtil.h"
#include "GenerationSettings.h"
#include "BrowserView.h"
#include "strutil.h"

OperationDialog::OperationDialog(OperationData * o, const QPoint &)
    : MyTabDialog(o->browser_node), oper(o),
      cl((ClassData *) ((BrowserClass *) o->browser_node->parent())->get_data()) {
  setCaption("Operation dialog");

  bool visit = !hasOkButton();
  
  QString stereotype = cl->get_stereotype();
  bool undef = (stereotype == "enum") || (stereotype == "typedef");
  QString st;
  
  st = GenerationSettings::cpp_class_stereotype(stereotype);
  cpp_undef = undef || (st == "enum") || (st == "typedef");
  
  st = GenerationSettings::java_class_stereotype(stereotype);
  java_undef = undef || (st == "enum") || (st == "typedef");
  
  st = GenerationSettings::idl_class_stereotype(stereotype);
  idl_undef = undef || (st == "enum") || (st == "typedef");
  
  QGrid * grid;
  QHBox * htab;
  QButtonGroup * bg;
    
  // general tab
  
  grid = new QGrid(2, this);
  grid->setSpacing(5);
  grid->setMargin(5);
  
  new QLabel("class : ", grid);
  new QLabel(((BrowserNode *) o->get_browser_node()->parent())->full_name(TRUE),
	     grid);
  
  new QLabel("name : ", grid);
  edname = new LineEdit(o->name(), grid);
  edname->setReadOnly(visit);
  
  new QLabel("value type : ", grid);
  
  nodes = BrowserClass::instances();
  nodes.full_names(list);

  edreturn_type = new QComboBox(!visit, grid);
  edreturn_type->insertItem(o->get_return_type().get_full_type());
  if (o->is_get_or_set)
    edreturn_type->setEnabled(FALSE);
  else if (! visit) {
    edreturn_type->insertStringList(GenerationSettings::basic_types());
    edreturn_type->insertStringList(list);
    edreturn_type->setAutoCompletion(TRUE);
  }
  edreturn_type->setCurrentItem(0);
  
  QSizePolicy sp = edreturn_type->sizePolicy();
  
  sp.setHorData(QSizePolicy::Expanding);
  edreturn_type->setSizePolicy(sp);

  new QLabel("stereotype : ", grid);
  edstereotype = new QComboBox(!visit, grid);
  edstereotype->insertItem(oper->stereotype);
  if (o->is_get_or_set)
    edstereotype->setEnabled(FALSE);
  else if (! visit) {
    edstereotype->insertStringList(BrowserOperation::default_stereotypes());
    edstereotype->setAutoCompletion(TRUE);
  }
  edstereotype->setCurrentItem(0);
  edstereotype->setSizePolicy(sp);

  new QLabel(grid);
  
  htab = new QHBox(grid);
  bg = uml_visibility.init(htab, o->get_uml_visibility());
  if (visit)
    bg->setEnabled(FALSE);
  htab->setStretchFactor(bg, 1000);
  
  htab->setStretchFactor(new QLabel("      ", htab), 0);
  
  bg = new QButtonGroup(2, QGroupBox::Horizontal, QString::null, htab);
  htab->setStretchFactor(bg, 1000);
  classoperation_cb = new QCheckBox("class operation", bg);
  classoperation_cb->setDisabled(visit);
  abstract_cb = new QCheckBox("abstract", bg);
  abstract_cb->setDisabled(visit);
  if (o->get_isa_class_operation())
    classoperation_cb->setChecked(TRUE);
  else if (o->get_is_abstract())
    abstract_cb->setChecked(TRUE);
  if (o->is_get_or_set)
    bg->setEnabled(FALSE);
  else if (! visit) {
    connect(classoperation_cb, SIGNAL(toggled(bool)),
	    SLOT(classoper_toggled(bool)));
    connect(abstract_cb, SIGNAL(toggled(bool)),
	    SLOT(abstract_toggled(bool)));
  }
  
  new QLabel("parameters : ", grid);
  table = new ParamsTable(o, grid, list, visit);
  if (o->is_get_or_set)
    table->setEnabled(FALSE);
  
  new QLabel("exceptions : ", grid);
  etable = new ExceptionsTable(o, grid, list, visit);
  
  QVBox * vtab = new QVBox(grid);
  new QLabel("description :", vtab);
  if (! visit)
    connect(new SmallPushButton("Editor", vtab), SIGNAL(clicked()),
	    this, SLOT(edit_description()));
  comment = new MultiLineEdit(grid);
  comment->setText(o->get_browser_node()->get_comment());
  QFont font = comment->font();
  font.setFamily("Courier");
  font.setFixedPitch(TRUE);
  comment->setFont(font);
  comment->setReadOnly(visit);
  
  addTab(grid, "Uml");
  
  // C++
  
  if (! cpp_undef) {
    grid = new QGrid(2, this);
    grid->setMargin(5);
    grid->setSpacing(5);
    
    new QLabel(grid);
    htab = new QHBox(grid);
    
    bg = cpp_visibility.init(htab, o->get_cpp_visibility(), 0, "follow uml");
    if (visit)
      bg->setEnabled(FALSE);
    
    htab->setStretchFactor(new QLabel("      ", htab), 0);
    
    bg = new QButtonGroup(5, QGroupBox::Horizontal, QString::null, htab);
    const_cb = new QCheckBox("const", bg);
    volatile_cb = new QCheckBox("volatile", bg);
    friend_cb = new QCheckBox("friend", bg);
    virtual_cb = new QCheckBox("virtual", bg);
    inline_cb = new QCheckBox("inline", bg);
    if (o->get_cpp_const())
      const_cb->setChecked(TRUE);
    if (o->get_is_volatile())
      volatile_cb->setChecked(TRUE);
    if (o->get_cpp_friend())
      friend_cb->setChecked(TRUE);
    if (o->get_cpp_virtual())
      virtual_cb->setChecked(TRUE);
    if (o->get_cpp_inline())
      inline_cb->setChecked(TRUE);
    if (visit) {
      const_cb->setDisabled(TRUE);
      volatile_cb->setDisabled(TRUE);
      virtual_cb->setDisabled(TRUE);
      inline_cb->setDisabled(TRUE);
    }
    else {
      connect(const_cb, SIGNAL(toggled(bool)),
	      SLOT(const_volatile_toggled(bool)));
      connect(volatile_cb, SIGNAL(toggled(bool)),
	      SLOT(const_volatile_toggled(bool)));
      connect(friend_cb, SIGNAL(toggled(bool)),
	      SLOT(friend_toggled(bool)));
      connect(virtual_cb, SIGNAL(toggled(bool)),
	      SLOT(virtual_toggled(bool)));
      connect(inline_cb, SIGNAL(toggled(bool)),
	      SLOT(inline_toggled(bool)));
    }
    
    if (o->is_get_or_set) {
      new QLabel("Name form : ", grid);
      edcppnamespec = new LineEdit(grid);
      edcppnamespec->setText(o->cpp_name_spec);
      if (visit)
	edcppnamespec->setReadOnly(TRUE);
      else {
	connect(edcppnamespec, SIGNAL(textChanged(const QString &)), this, SLOT(cpp_update_decl()));
	connect(edcppnamespec, SIGNAL(textChanged(const QString &)), this, SLOT(cpp_update_def()));
      }
    }
    else
      edcppnamespec = 0;
    
    new QLabel("Declaration : ", grid);
    edcppdecl = new MultiLineEdit(grid);
    edcppdecl->setText(o->get_cppdecl());
    edcppdecl->setFont(font);
    if (visit)
      edcppdecl->setReadOnly(TRUE);
    else
      connect(edcppdecl, SIGNAL(textChanged()), this, SLOT(cpp_update_decl()));
    
    new QLabel("Result after\nsubstitution : ", grid);
    showcppdecl = new MultiLineEdit(grid);
    showcppdecl->setReadOnly(TRUE);
    showcppdecl->setFont(font);
    
    if (! visit) {
      new QLabel(grid);
      htab = new QHBox(grid);  
      if (! o->is_get_or_set) {
	connect(new QPushButton("Default declaration", htab), SIGNAL(pressed ()),
		this, SLOT(cpp_default_decl()));
	connect(new QPushButton("From definition", htab), SIGNAL(pressed ()),
		this, SLOT(cpp_decl_from_def()));
      }
      connect(new QPushButton("Not generated in C++", htab), SIGNAL(pressed ()),
	      this, SLOT(cpp_unmapped_decl()));
    }
    
    new QLabel("Definition :", grid);
    edcppdef = new MultiLineEdit(grid);
    edcppdef->setText(o->get_cppdef());
    edcppdef->setFont(font);
    if (visit)
      edcppdef->setReadOnly(TRUE);
    else
      connect(edcppdef, SIGNAL(textChanged()), this, SLOT(cpp_update_def()));
    
    new QLabel("Result after\nsubstitution : ", grid);
    showcppdef = new MultiLineEdit(grid);
    showcppdef->setReadOnly(TRUE);
    showcppdef->setFont(font);
    
    editcppbody = new SmallPushButton((visit) ? "Show body" : "Edit body", grid);
    connect(editcppbody, SIGNAL(clicked()), this, SLOT(cpp_edit_body()));
    
    char * b = o->get_body(TRUE);
    
    if (b != 0) {
      cppbody = oldcppbody = b;
      delete [] b;
    }
        
    if (! visit) {
      htab = new QHBox(grid);  
      if (! o->is_get_or_set) {
	connect(new QPushButton("Default definition", htab), SIGNAL(pressed ()),
		this, SLOT(cpp_default_def()));
	connect(new QPushButton("From declaration", htab), SIGNAL(pressed ()),
		this, SLOT(cpp_def_from_decl()));
      }
      connect(new QPushButton("Not generated in C++", htab), SIGNAL(pressed ()),
	      this, SLOT(cpp_unmapped_def()));
    }
    
    addTab(grid, "C++");
    
    cl->get_class_spec(templates, cl_names);
  }
  
  // Java
  
  if (! java_undef) {
    grid = new QGrid(2, this);
    grid->setMargin(5);
    grid->setSpacing(5);
    
    new QLabel(grid);
    bg = new QButtonGroup(2, QGroupBox::Horizontal, QString::null, grid);
    final_cb = new QCheckBox("final", bg);
    if (o->get_java_final())
      final_cb->setChecked(TRUE);
    if (visit)
      final_cb->setDisabled(TRUE);
    else
      connect(final_cb, SIGNAL(toggled(bool)),
	      SLOT(finalsynchronized_toggled(bool)));

    synchronized_cb = new QCheckBox("synchronized", bg);
    if (o->get_java_synchronized())
      synchronized_cb->setChecked(TRUE);
    if (visit)
      synchronized_cb->setDisabled(TRUE);
    else
      connect(synchronized_cb, SIGNAL(toggled(bool)),
	      SLOT(finalsynchronized_toggled(bool)));
    
    if (o->is_get_or_set) {
      new QLabel("Name form : ", grid);
      edjavanamespec = new LineEdit(grid);
      edjavanamespec->setText(o->java_name_spec);
      if (visit)
	edjavanamespec->setDisabled(TRUE);
      else
	connect(edjavanamespec, SIGNAL(textChanged(const QString &)), this, SLOT(java_update_def()));
    }
    else
      edjavanamespec = 0;
    
    new QLabel("Definition :", grid);
    edjavadef = new MultiLineEdit(grid);
    edjavadef->setText(o->get_javadef());
    edjavadef->setFont(font);
    if (visit)
      edjavadef->setReadOnly(TRUE);
    else
      connect(edjavadef, SIGNAL(textChanged()), this, SLOT(java_update_def()));
    
    new QLabel("Result after\nsubstitution : ", grid);
    showjavadef = new MultiLineEdit(grid);
    showjavadef->setReadOnly(TRUE);
    showjavadef->setFont(font);
    
    editjavabody = new SmallPushButton((visit) ? "Show body" : "Edit body", grid);
    connect(editjavabody, SIGNAL(clicked()), this, SLOT(java_edit_body()));
    
    char * b = o->get_body(FALSE);
    
    if (b != 0) {
      javabody = oldjavabody = b;
      delete [] b;
    }
        
    if (! visit) {
      htab = new QHBox(grid);  
      if (! o->is_get_or_set)
	connect(new QPushButton("Default definition", htab), SIGNAL(pressed ()),
		this, SLOT(java_default_def()));
      connect(new QPushButton("Not generated in Java", htab), SIGNAL(pressed ()),
	      this, SLOT(java_unmapped_def()));
    }
    
    addTab(grid, "Java");
  }
    
  // IDL
  
  if (! idl_undef) {
    grid = new QGrid(2, this);
    grid->setMargin(5);
    grid->setSpacing(5);
    
    new QLabel(grid);
    bg = new QButtonGroup(1, QGroupBox::Horizontal, QString::null, grid);
    oneway_cb = new QCheckBox("oneway", bg);
    if (o->get_idl_oneway())
      oneway_cb->setChecked(TRUE);
    if (visit)
      oneway_cb->setDisabled(TRUE);
    else
      connect(oneway_cb, SIGNAL(toggled(bool)),
	      SLOT(oneway_toggled(bool)));
    
    if (o->is_get_or_set) {
      new QLabel("Name form : ", grid);
      edidlnamespec = new LineEdit(grid);
      edidlnamespec->setText(o->idl_name_spec);
      if (visit)
	edidlnamespec->setReadOnly(TRUE);
      else
	connect(edidlnamespec, SIGNAL(textChanged(const QString &)), this, SLOT(idl_update_decl()));
    }
    else
      edidlnamespec = 0;
    
    new QLabel("Declaration :", grid);
    edidldecl = new MultiLineEdit(grid);
    edidldecl->setText(o->get_idldecl());
    edidldecl->setFont(font);
    if (visit)
      edidldecl->setReadOnly(TRUE);
    else
      connect(edidldecl, SIGNAL(textChanged()), this, SLOT(idl_update_decl()));
    
    new QLabel("Result after\nsubstitution : ", grid);
    showidldecl = new MultiLineEdit(grid);
    showidldecl->setReadOnly(TRUE);
    showidldecl->setFont(font);
    
    if (! visit) {
      new QLabel(grid);
      htab = new QHBox(grid);  
      if (! o->is_get_or_set)
	connect(new QPushButton("Default declaration", htab), SIGNAL(pressed ()),
		this, SLOT(idl_default_def()));
      connect(new QPushButton("Not generated in Idl", htab), SIGNAL(pressed ()),
	      this, SLOT(idl_unmapped_def()));
    }
    
    addTab(grid, "Idl");
  }
  
  // USER : list key - value
  
  grid = new QGrid(2, this);
  grid->setMargin(5);
  grid->setSpacing(5);
  
  kvtable = new KeyValuesTable(o->get_browser_node(), grid, visit);
  addTab(grid, "User");
  
  //
  
  connect(this, SIGNAL(currentChanged(QWidget *)),
	  this, SLOT(update_all_tabs()));
}

void OperationDialog::edit_description() {
  edit(comment->text(), edname->text().stripWhiteSpace() + "_description",
       oper, TxtEdit, this, (post_edit) post_edit_description);
}

void OperationDialog::post_edit_description(OperationDialog * d, QString s)
{
  d->comment->setText(s);
}

void OperationDialog::accept() {
  BrowserNode * bn = oper->browser_node;
  QString s = edname->text().stripWhiteSpace();
  
  if ((s != oper->name()) &&
      ((BrowserNode *) bn->parent())->wrong_child_name(s, UmlOperation,
						       bn->allow_spaces(),
						       bn->allow_empty()))
    QMessageBox::critical(0, "Error", s + "\n\nillegal name or already used");
  else {  
    bn->set_name(s);
    
    if (!oper->is_get_or_set) {
      AType t;
      
      s = edreturn_type->currentText().stripWhiteSpace();
      
      if (! s.isEmpty()) {
	int index = list.findIndex(edreturn_type->currentText());
	
	if (index >= 0)
	  t.type = (BrowserClass *) nodes.at(index);
	else
	  t.explicit_type = s;
      }
      oper->set_return_type(t);
      
      oper->set_stereotype(edstereotype->currentText().stripWhiteSpace());
    }
    
    oper->uml_visibility = uml_visibility.value();
    
    oper->isa_class_operation = classoperation_cb->isChecked();
    oper->set_is_abstract(abstract_cb->isChecked());
    
    table->update(oper, nodes);
    etable->update(oper, nodes);
    
    bn->set_comment(comment->text());
    UmlWindow::set_commented(bn);
    
    // C++
    
    if (cpp_undef) {
      oper->cpp_decl = QString::null;
      oper->cpp_def.assign(QString::null, TRUE);
      if (!oldcppbody.isEmpty())
	oper->new_body(QString::null, TRUE);
    }
    else {
      if (oper->is_get_or_set) {
	oper->cpp_name_spec = edcppnamespec->text().stripWhiteSpace();
	oper->cpp_body.length = 0;
      }
      else if (!abstract_cb->isChecked() && 
	       (edcppdef->text().find("${body}") != -1)) {
	if (cppbody != oldcppbody)
	  oper->new_body(cppbody, TRUE);
      }
      else if(!oldcppbody.isEmpty())
	oper->new_body(QString::null, TRUE);
      oper->cpp_decl = edcppdecl->text();
      oper->cpp_visibility = cpp_visibility.value();
      
      oper->cpp_const = const_cb->isChecked();
      oper->is_volatile = volatile_cb->isChecked();
      oper->cpp_friend = friend_cb->isChecked();
      oper->cpp_virtual = virtual_cb->isChecked();
      oper->cpp_inline = inline_cb->isChecked();
      oper->cpp_def.assign(edcppdef->text(),
			   abstract_cb->isChecked() ||
			   (edcppdef->text().find("${body}") != -1));
    }
    
    // java
    
    if (java_undef) {
      oper->java_def.assign(QString::null, TRUE);
      if (!oldjavabody.isEmpty())
	oper->new_body(QString::null, FALSE);
    }
    else {
      if (oper->is_get_or_set)
	oper->java_name_spec = edjavanamespec->text().stripWhiteSpace();
      else if (!abstract_cb->isChecked() && 
	       strcmp(GenerationSettings::java_class_stereotype(cl->get_stereotype()),
		      "interface") &&
	       (edjavadef->text().find("${body}") != -1)) {
	if (javabody != oldjavabody)
	  oper->new_body(javabody, FALSE);
      }
      else if(!oldjavabody.isEmpty())
	oper->new_body(QString::null, FALSE);
      oper->java_final = final_cb->isChecked();
      oper->java_synchronized = synchronized_cb->isChecked();
      oper->java_def.assign(edjavadef->text(),
			    abstract_cb->isChecked() || 
			    (GenerationSettings::java_class_stereotype(cl->get_stereotype())
			     == "interface") ||
			    (edjavadef->text().find("${body}") != -1));
    }
    
    // idl
    
    if (idl_undef)
      oper->idl_decl = QString::null;
    else {
      if (oper->is_get_or_set)
	oper->idl_name_spec = edidlnamespec->text().stripWhiteSpace();
      oper->idl_oneway = oneway_cb->isChecked();
      oper->idl_decl = edidldecl->text();
    }
    
    // user
    
    kvtable->update(bn);
    
    bn->modified();
    bn->package_modified();
    oper->modified();
    
    MyTabDialog::accept();
  }
}

void OperationDialog::classoper_toggled(bool on) {
  if (on)
    abstract_cb->setChecked(FALSE);
}

void OperationDialog::abstract_toggled(bool on) {
  if (on)
    classoperation_cb->setChecked(FALSE);
}

void OperationDialog::update_all_tabs() {
  if (! cpp_undef) {
    cpp_update_decl();
    cpp_update_def();
  }
  if (! java_undef)
    java_update_def();
  if (! idl_undef)
    idl_update_decl();
}

// C++

void OperationDialog::const_volatile_toggled(bool) {
  cpp_update_decl();
  cpp_update_def();
}

void OperationDialog::virtual_toggled(bool on) {
  if (on) {
    friend_cb->setChecked(FALSE);
    inline_cb->setChecked(FALSE);
  }
  cpp_update_decl();
  cpp_update_def();
}

void OperationDialog::friend_toggled(bool on) {
  if (on) {
    inline_cb->setChecked(FALSE);
    virtual_cb->setChecked(FALSE);
  }
  cpp_update_decl();
  cpp_update_def();
}

void OperationDialog::inline_toggled(bool on) {
  if (on) {
    friend_cb->setChecked(FALSE);
    virtual_cb->setChecked(FALSE);
  }
  cpp_update_decl();
  cpp_update_def();
}

void OperationDialog::cpp_default_decl() {
  QString s = oper->default_cpp_decl(edname->text().stripWhiteSpace());
  int index;
  QString params;
  int nparams = (int) table->nparams();
  QString sep;
  
  for (index = 0; index != nparams; index += 1, sep = ", ")
    params += sep + GenerationSettings::cpp(the_type(table->type(index),
						     list, nodes),
					    table->dir(index), index);
  
  if ((index = s.find("${)}")) != -1)
    s.insert(index, params);
  
  edcppdecl->setText(s);
  cpp_update_decl();
}

void OperationDialog::cpp_unmapped_decl() {
  edcppdecl->setText(QString::null);
  showcppdecl->setText(QString::null);

  cpp_unmapped_def();
}

void OperationDialog::cpp_decl_from_def() {
  QString dcl = oper->default_cpp_decl(edname->text().stripWhiteSpace());
  QString def = edcppdef->text();
  int index1;
  int index2;
  int index3;
  int index4;
  
  // copy args
  if (((index1 = def.find("${(}") + 4) == (-1 + 4)) ||
      ((index2 = def.find("${)}", index1)) == -1) ||
      ((index3 = dcl.find("${(}") + 4) == (-1 + 4)) ||
      ((index4 = dcl.find("${)}", index3)) == -1))
      // wrong specification(s)
    return;
  
  if (index4 != index3)
    dcl.remove(index3, index4 - index3);
  if (index2 != index1)
    dcl.insert(index3, def.mid(index1, index2 - index1));

  // copy return type modifications
  if (((index1 = def.find("${type}")) != -1) &&
      ((index2 = def.find("${class}", index1 + 7)) != -1) &&
      ((index3 = dcl.find("${type}")) != -1) &&
      ((index4 = dcl.find("${name}", index3 + 7)) != -1)) {
    // valid specification
    while ((index1 != 0) && (def.at(index1 - 1) != QChar('}')))
      index1 -= 1;
    while ((index3 != 0) && (dcl.at(index3 - 1) != QChar('}')))
      index3 -= 1;
    dcl.remove(index3, index4 - index3);
    dcl.insert(index3, def.mid(index1, index2 - index1));
  }

  // manage const
  if ((def.find("${const}") == -1) && // const removed
      ((index1 = dcl.find("${const}")) != -1))
    // remove const
    dcl.remove(index1, 8);

  // manage throw
  if ((def.find("${throw}") == -1) && // throw removed
      ((index1 = dcl.find("${throw}")) != -1))
    // remove throw
    dcl.remove(index1, 8);

  // update decl
  edcppdecl->setText(dcl);
  cpp_update_decl();
}

void OperationDialog::manage_cpp_type(unsigned rank, QString & s)
{
  if (rank < table->nparams()) 
    s += GenerationSettings::cpp_type(type(table->type(rank), list, nodes));
  else {
    s += "${t";
    s += QString::number(rank);
    s += '}';
  }
}

void OperationDialog::manage_var(unsigned rank, QString & s)
{
  if (rank < table->nparams()) 
    s += table->name(rank);
  else {
    s += "${p";
    s += QString::number(rank);
    s += '}';
  }
}

QString OperationDialog::compute_name(LineEdit * spec) {
  if (spec != 0) {
    QString s = spec->text().stripWhiteSpace();
    int index;
    
    if ((index = s.find("${name}")) != -1)
      s.replace(index, 7, ((BrowserOperation *) oper->browser_node)->get_of_name());
    else if ((index = s.find("${Name}")) != -1)
      s.replace(index, 7,
		capitalize(((BrowserOperation *) oper->browser_node)->get_of_name()));
    return s;
  }
  else
    return edname->text();
}

void OperationDialog::manage_cpp_exceptions(QString & s)
{
  int index;
  int ne = etable->nexceptions();
  const char * sep = " throw (";
	
  for (index = 0; index != ne; index += 1) {
    s += sep;
    s += GenerationSettings::cpp_type(type(etable->type(index), list, nodes));
    sep = ", ";
  }
  
  if (index != 0)
    s += ')';
  else if (GenerationSettings::cpp_oper_force_throw())
    s += " throw ()";
}

void OperationDialog::cpp_update_decl() {
  cpp_visibility.update_default(uml_visibility);
  
  // do NOT write
  //	const char * p = edcppdecl->text();
  // because the QString is immediatly destroyed !
  QString def = edcppdecl->text();
  const char * p = def;
  QString indent = "";
  QString s;
  unsigned rank;

  while ((*p == ' ') || (*p == '\t'))
    indent += *p++;
  
  if (*p != '#')
    s = indent;
  
  while (*p) {
    if (!strncmp(p, "${comment}", 10))
      manage_comment(comment->text(), p, s, indent);
    else if (!strncmp(p, "${friend}", 9)) {
      p += 9;
      if (friend_cb->isChecked())
	s += "friend ";
    }
    else if (!strncmp(p, "${static}", 9)) {
      p += 9;
      if (classoperation_cb->isChecked())
	s += "static ";
    }
    else if (!strncmp(p, "${virtual}", 10)) {
      p += 10;
      if (virtual_cb->isChecked())
	s += "virtual ";
    }
    else if (!strncmp(p, "${type}", 7)) {
      p += 7;
      s += GenerationSettings::cpp_type(type(edreturn_type->currentText().stripWhiteSpace(),
					     list, nodes));
    }
    else if (!strncmp(p, "${name}", 7)) {
      p += 7;
      s += compute_name(edcppnamespec);
    }
    else if (!strncmp(p, "${(}", 4)) {
      p += 4;
      s += '(';
    }
    else if (!strncmp(p, "${)}", 4)) {
      p += 4;
      s += ')';
    }
    else if (!strncmp(p, "${const}", 8)) {
      p += 8;
      if (const_cb->isChecked())
	s += " const";
    }
    else if (!strncmp(p, "${volatile}", 11)) {
      p += 11;
      if (volatile_cb->isChecked())
	s += " volatile";
    }
    else if (!strncmp(p, "${throw}", 8)) {
      p += 8;
      manage_cpp_exceptions(s);
    }
    else if (!strncmp(p, "${abstract}", 11)) {
      p += 11;
      if (abstract_cb->isChecked())
	s += " = 0";
    }
    else if (sscanf(p, "${t%u}", &rank) == 1) {
      manage_cpp_type(rank, s);
      p = strchr(p, '}') + 1;
    }
    else if (sscanf(p, "${p%u}", &rank) == 1) {
      manage_var(rank, s);
      p = strchr(p, '}') + 1;
    }
    else if (oper->is_get_or_set && !strncmp(p, "${stereotype}", 13)) {
      p += 13;
      // get/set relation with multiplicity > 1
      s += GenerationSettings::cpp_relation_stereotype(oper->stereotype);
    }
    else if (*p == '\n') {
      s += *p++;
      if (*p && (*p != '#'))
	s += indent;
    }
    else
      s += *p++;
  }

  showcppdecl->setText(s);
}

QString OperationDialog::cpp_decl(const BrowserOperation * op)
{
  OperationData * d = (OperationData *) op->get_data();
  QCString decl = d->cpp_decl;
  
  remove_comments(decl);
  remove_preprocessor(decl);
  
  const char * p = decl;
  QString s;
  unsigned rank;

  while ((*p == ' ') || (*p == '\t'))
    p += 1;
  
  while (*p) {
    if (!strncmp(p, "${comment}", 10))
      p += 10;
    else if (!strncmp(p, "${friend}", 9))
      p += 9;
    else if (!strncmp(p, "${static}", 9))
      p += 9;
    else if (!strncmp(p, "${virtual}", 10))
      p += 10;
    else if (!strncmp(p, "${type}", 7)) {
      p += 7;
      s += GenerationSettings::cpp_type(d->return_type.get_type());
    }
    else if (!strncmp(p, "${name}", 7)) {
      p += 7;
      s += op->compute_name(d->cpp_name_spec);
    }
    else if (!strncmp(p, "${(}", 4)) {
      p += 4;
      s += '(';
    }
    else if (!strncmp(p, "${)}", 4)) {
      p += 4;
      s += ')';
    }
    else if (!strncmp(p, "${const}", 8))
      p += 8;
    else if (!strncmp(p, "${volatile}", 11))
      p += 11;
    else if (!strncmp(p, "${throw}", 8))
      p += 8;
    else if (!strncmp(p, "${abstract}", 11)) 
      break;
    else if (sscanf(p, "${t%u}", &rank) == 1) {
      if (rank < d->nparams) 
	s += GenerationSettings::cpp_type(d->params[rank].get_type().get_type());
      else {
	s += "${t";
	s += QString::number(rank);
	s += '}';
      }
      p = strchr(p, '}') + 1;
    }
    else if (sscanf(p, "${p%u}", &rank) == 1) {
      if (rank < d->nparams) 
	s += d->params[rank].get_name();
      else {
	s += "${p";
	s += QString::number(rank);
	s += '}';
      }
      p = strchr(p, '}') + 1;
    }
    else if (d->is_get_or_set && !strncmp(p, "${stereotype}", 13)) {
      p += 13;
      // get/set relation with multiplicity > 1
      s += GenerationSettings::cpp_relation_stereotype(d->stereotype);
    }
    else if (*p == '\n') {
      s += ' ';
      do
	p+= 1;
      while ((*p == ' ') || (*p == '\t'));
    }
    else if ((*p == '{') || (*p == ';'))
      break;
    else
      s += *p++;
  }

  return s;
}

void OperationDialog::cpp_default_def() {
  QString s = oper->default_cpp_def(edname->text().stripWhiteSpace());
  QString params;
  int index;
  int nparams = (int) table->nparams();
  QString sep;
  
  for (index = 0; index != nparams; index += 1, sep = ", ")
    params += sep + GenerationSettings::cpp(the_type(table->type(index),
						     list, nodes),
					    table->dir(index), index);
  
  if ((index = s.find("${)}")) != -1)
    s.insert(index, params);
  
  edcppdef->setText(s);
  cpp_update_def();
}

void OperationDialog::cpp_unmapped_def() {
  edcppdef->setText(QString::null);
  showcppdef->setText(QString::null);
}

void OperationDialog::cpp_def_from_decl() {
  
  QString dcl = edcppdecl->text();
  QString def;
  
  // manage abstract
  if ((dcl.find("${abstract}") == -1) || // abstract removed
      !abstract_cb->isChecked()) {
    // probably not abstract
    int index1;
    int index2;
    int index3;
    int index4;
    
    def = oper->default_cpp_def(edname->text().stripWhiteSpace());
    
    // copy args
    if (((index1 = dcl.find("${(}") + 4) == (-1 + 4)) ||
	((index2 = dcl.find("${)}", index1)) == -1) ||
	((index3 = def.find("${(}") + 4) == (-1 + 4)) ||
	((index4 = def.find("${)}", index3)) == -1))
      // wrong specification(s)
      return;
    
    if (index4 != index3)
      def.remove(index3, index4 - index3);
    if (index2 != index1)
      def.insert(index3, dcl.mid(index1, index2 - index1));
    
    // copy return type modifications
    if (((index1 = dcl.find("${type}")) != -1) &&
	((index2 = dcl.find("${name}", index1 + 7)) != -1) &&
	((index3 = def.find("${type}")) != -1) &&
	((index4 = def.find("${class}", index3 + 7)) != -1)) {
      // valid specifications
      while ((index1 != 0) && (dcl.at(index1 - 1) != QChar('}')))
	index1 -= 1;
      while ((index3 != 0) && (def.at(index3 - 1) != QChar('}')))
	index3 -= 1;
      def.remove(index3, index4 - index3);
      def.insert(index3, dcl.mid(index1, index2 - index1));
    }
    
    // manage const
    if ((dcl.find("${const}") == -1) && // const removed
	((index1 = def.find("${const}")) != -1))
      // remove const
      def.remove(index1, 8);
    
    // manage throw
    if ((dcl.find("${throw}") == -1) && // throws removed
	((index1 = def.find("${throw}")) != -1))
      // remove throw
      def.remove(index1, 8);
  }
  
  // update def
  edcppdef->setText(def);
  cpp_update_def();
}

static void insert_template(const QString & tm, const char *& p, 
			    QString & s, const QString & indent)
{
  // search the beginning of the definition/declaration in p;
  for (;;) {
    if (*p == 0)
      return;
    
    if (*p == '\n') {
      s += *p++;
      if (*p != '#')
	s += indent;
    }
    if (*p <= ' ')
      s += *p++;
    else if (*p == '/') {
      if (p[1] == '/') {
	// comment
	s += *p++;
	do s += *p++;
	while (*p && (*p != '\n'));
      }
      else if (p[1] == '*') {
	/* comment */
	s += *p++;
	do s += *p++;
	while (*p && ((*p != '*') || (p[1] != '/')));
	s += "*/";
	p += 2;
      }
      else
	break;
    }
    else if (*p == '#') {
      do {
	s += *p++;
	if (*p == '\\') {
	  s += *p++;
	  s += *p++;
	}
	else if ((*p == '/') && (p[1] == '*')) {
	  /* comment */
	  s += *p++;
	  do s += *p++;
	  while (*p && ((*p != '*') || (p[1] != '/')));
	  s += "*/";
	  p += 2;
	}
      } while (*p && (*p != '\n'));
    }
    else
      break;
  }
  
  s += tm;
}

void OperationDialog::cpp_update_def() {
  QString def = edcppdef->text();
  QString s;
  
  if (!abstract_cb->isChecked()) {
    // do NOT write
    //	const char * p = edcppdef->text()
    // because the QString is immediatly destroyed !
    const char * p = def;
    QString indent = "";
    unsigned rank;
    
    while ((*p == ' ') || (*p == '\t'))
      indent += *p++;
    
    if (! templates.isEmpty())
      insert_template(templates, p, s, indent);
    
    if (*p != '#')
      s += indent;
  
    while (*p) {
      if (!strncmp(p, "${comment}", 10))
	manage_comment(comment->text(), p, s, indent);
      else if (!strncmp(p, "${inline}", 9)) {
	p += 9;
	if (inline_cb->isChecked())
	  s += "inline ";
      }
      else if (!strncmp(p, "${type}", 7)) {
	p += 7;
	s += GenerationSettings::cpp_type(type(edreturn_type->currentText().stripWhiteSpace(),
					       list, nodes));
      }
      else if (!strncmp(p, "${class}", 8)) {
	p += 8;
	if (friend_cb->isChecked() && !strncmp(p, "::", 2))
	  p += 2;
	else
	  s += cl_names;
      }
      else if (!strncmp(p, "${name}", 7)) {
	p += 7;
	s += compute_name(edcppnamespec);
      }
      else if (!strncmp(p, "${(}", 4)) {
	p += 4;
	s += '(';
      }
      else if (!strncmp(p, "${)}", 4)) {
	p += 4;
	s += ')';
      }
      else if (!strncmp(p, "${const}", 8)) {
	p += 8;
	if (const_cb->isChecked())
	  s += " const";
      }
      else if (!strncmp(p, "${volatile}", 11)) {
	p += 11;
	if (volatile_cb->isChecked())
	  s += " volatile";
      }
      else if (!strncmp(p, "${throw}", 8)) {
	p += 8;
      manage_cpp_exceptions(s);
      }
      else if (!strncmp(p, "${staticnl}", 11)) {
	p += 11;
	if (classoperation_cb->isChecked()) {
	  s += '\n';
	  s += indent;
	}
	else
	  s += ' ';
      }
      else if (sscanf(p, "${t%u}", &rank) == 1) {
	manage_cpp_type(rank, s);
	p = strchr(p, '}') + 1;
      }
      else if (sscanf(p, "${p%u}", &rank) == 1) {
	manage_var(rank, s);
	p = strchr(p, '}') + 1;
      }
      else if (oper->is_get_or_set && !strncmp(p, "${stereotype}", 13)) {
	p += 13;
	// get/set relation with multiplicity > 1
	s += GenerationSettings::cpp_relation_stereotype(oper->stereotype);
      }
      else if (*p == '\n') {
	s += *p++;
	if (*p && (*p != '#'))
	  s += indent;
      }
      else
	s += *p++;
    }
    
    editcppbody->setEnabled(def.find("${body}") != -1);
  }
  else
    editcppbody->setEnabled(FALSE);
  
  showcppdef->setText(s);
}

void OperationDialog::cpp_edit_body() {
  edit(cppbody, edname->text().stripWhiteSpace() + "_body",
       oper, CppEdit, this, (post_edit) post_cpp_edit_body);
}

void OperationDialog::post_cpp_edit_body(OperationDialog * d, QString s)
{
  d->cppbody = s;
}

// Java

void OperationDialog::finalsynchronized_toggled(bool) {
  java_update_def();
}

void OperationDialog::manage_java_type(unsigned rank, QString & s)
{
  if (rank < table->nparams()) 
    s += GenerationSettings::java_type(type(table->type(rank), list, nodes));
  else {
    s += "${t";
    s += QString::number(rank);
    s += '}';
  }
}

void OperationDialog::manage_java_exceptions(QString & s)
{
  int index;
  int ne = etable->nexceptions();
  const char * sep = " throws ";
	
  for (index = 0; index != ne; index += 1) {
    s += sep;
    s += GenerationSettings::java_type(type(etable->type(index), list, nodes));
    sep = ", ";
  }
  
  if (index != 0)
    s += ' ';
}

void OperationDialog::java_default_def() {
  QString s = oper->default_java_def(edname->text().stripWhiteSpace());
  QString params;
  int index;
  int nparams = (int) table->nparams();
  const char * sep;
  
  for (index = 0, sep = ""; index != nparams; index += 1, sep = ", ") {
    QString p;
    
    p.sprintf("%s${t%d} ${p%d}", sep, index, index);
    params += p;
  }
  
  if ((index = s.find("${)}")) != -1)
    s.insert(index, params);
  
  edjavadef->setText(s);
  java_update_def();
}

void OperationDialog::java_unmapped_def() {
  edjavadef->setText(QString::null);
  showjavadef->setText(QString::null);
}

void OperationDialog::java_update_def() {
  // do NOT write
  //	const char * p = edjavadef->text();
  // because the QString is immediatly destroyed !
  QString def = edjavadef->text();
  bool interf = (GenerationSettings::java_class_stereotype(cl->get_stereotype())
		 == "interface");
  bool nobody = (abstract_cb->isChecked() || interf);
  const char * p = def;
  QString indent = "";
  QString s;
  unsigned rank;
  
  while ((*p == ' ') || (*p == '\t'))
    indent += *p++;
  
  s = indent;
  
  while (*p) {
    if (!strncmp(p, "${comment}", 10))
      manage_comment(comment->text(), p, s, indent);
    else if (!strncmp(p, "${final}", 8)) {
      p += 8;
      if (final_cb->isChecked())
	s += "final ";
    }
    else if (!strncmp(p, "${visibility}", 13)) {
      p += 13;
      s += uml_visibility.state() + ' ';
    }
    else if (!strncmp(p, "${static}", 9)) {
      p += 9;
      if (classoperation_cb->isChecked())
	s += "static ";
    }
    else if (!strncmp(p, "${abstract}", 11)) {
      p += 11;
      if (abstract_cb->isChecked() && !interf)
	s += "abstract ";
    }
    else if (!strncmp(p, "${synchronized}", 15)) {
      p += 15;
      if (synchronized_cb->isChecked())
	s += "synchronized ";
    }
    else if (!strncmp(p, "${type}", 7)) {
      p += 7;
      s += GenerationSettings::java_type(type(edreturn_type->currentText().stripWhiteSpace(),
					      list, nodes));
    }
    else if (!strncmp(p, "${name}", 7)) {
      p += 7;
      s += compute_name(edjavanamespec);
    }
    else if (!strncmp(p, "${(}", 4)) {
      p += 4;
      s += '(';
    }
    else if (!strncmp(p, "${)}", 4)) {
      p += 4;
      s += ')';
    }
    else if (!strncmp(p, "${throws}", 9)) {
      manage_java_exceptions(s);
      p += 9;
    }
    else if (!strncmp(p, "${staticnl}", 11)) {
      p += 11;
      if (classoperation_cb->isChecked()) {
	s += '\n';
	s += indent;
      }
      else
	s += ' ';
    }
    else if (sscanf(p, "${t%u}", &rank) == 1) {
      manage_java_type(rank, s);
      p = strchr(p, '}') + 1;
    }
    else if (sscanf(p, "${p%u}", &rank) == 1) {
      manage_var(rank, s);
      p = strchr(p, '}') + 1;
    }
    else if (oper->is_get_or_set && !strncmp(p, "${stereotype}", 13)) {
      p += 13;
      // get/set relation with multiplicity > 1
      s += GenerationSettings::java_relation_stereotype(oper->stereotype);
    }
    else if (*p == '\n') {
      s += *p++;
      if (*p)
	s += indent;
    }
    else if ((*p == '{') && nobody) {
      s += ";";
      break;
    }
    else
      s += *p++;
  }
    
  editjavabody->setEnabled(!nobody && (def.find("${body}") != -1));
  
  showjavadef->setText(s);
}

QString OperationDialog::java_decl(const BrowserOperation * op)
{
  OperationData * d = (OperationData *) op->get_data();
  QCString decl = d->java_def;
  
  remove_comments(decl);
  
  const char * p = decl;
  QString s;
  unsigned rank;

  while ((*p == ' ') || (*p == '\t'))
    p += 1;
  
  while (*p) {
    if (!strncmp(p, "${comment}", 10))
      p += 10;
    else if (!strncmp(p, "${final}", 8))
      p += 8;
    else if (!strncmp(p, "${visibility}", 13))
      p += 13;
    else if (!strncmp(p, "${static}", 9))
      p += 9;
    else if (!strncmp(p, "${abstract}", 11))
      p += 11;
    else if (!strncmp(p, "${synchronized}", 15))
      p += 15;
    else if (!strncmp(p, "${type}", 7)) {
      p += 7;
      s += GenerationSettings::java_type(d->return_type.get_type());
    }
    else if (!strncmp(p, "${name}", 7)) {
      p += 7;
      s += op->compute_name(d->java_name_spec);
    }
    else if (!strncmp(p, "${(}", 4)) {
      p += 4;
      s += '(';
    }
    else if (!strncmp(p, "${)}", 4)) {
      p += 4;
      s += ')';
    }
    else if (!strncmp(p, "${throws}", 9))
      p += 9;
    else if (!strncmp(p, "${staticnl}", 11)) {
      p += 11;
      s += ' ';
    }
    else if (sscanf(p, "${t%u}", &rank) == 1) {
      if (rank < d->nparams) 
	s += GenerationSettings::java_type(d->params[rank].get_type().get_type());
      else {
	s += "${t";
	s += QString::number(rank);
	s += '}';
      }
      p = strchr(p, '}') + 1;
    }
    else if (sscanf(p, "${p%u}", &rank) == 1) {
      if (rank < d->nparams) 
	s += d->params[rank].get_name();
      else {
	s += "${p";
	s += QString::number(rank);
	s += '}';
      }
      p = strchr(p, '}') + 1;
    }
    else if (d->is_get_or_set && !strncmp(p, "${stereotype}", 13)) {
      p += 13;
      // get/set relation with multiplicity > 1
      s += GenerationSettings::java_relation_stereotype(d->stereotype);
    }
    else if (*p == '\n') {
      s += ' ';
      do
	p+= 1;
      while ((*p == ' ') || (*p == '\t'));
    }
    else if ((*p == '{') || (*p == ';'))
      break;
    else
      s += *p++;
  }
  
  return s;
}

void OperationDialog::java_edit_body() {
  edit(javabody, edname->text().stripWhiteSpace() + "_body",
       oper, JavaEdit, this, (post_edit) post_java_edit_body);
}

void OperationDialog::post_java_edit_body(OperationDialog * d, QString s)
{
  d->javabody = s;
}

// Idl

void OperationDialog::oneway_toggled(bool) {
  idl_update_decl();
}

void OperationDialog::manage_idl_type(unsigned rank, QString & s)
{
  if (rank < table->nparams()) 
    s += GenerationSettings::idl_type(type(table->type(rank), list, nodes));
  else {
    s += "${t";
    s += QString::number(rank);
    s += '}';
  }
}

void OperationDialog::manage_dir(unsigned rank, QString & s)
{
  if (rank < table->nparams()) {
    switch (table->dir(rank)) {
    case UmlIn:
      s += "in";
      break;
    case UmlOut:
      s += "out";
      break;
    default:
      s += "inout";
    }
  }
  else {
    s += "${d";
    s += QString::number(rank);
    s += '}';
  }
}

void OperationDialog::manage_idl_exceptions(QString & s)
{
  int index;
  int ne = etable->nexceptions();
  const char * sep = " raises (";
	
  for (index = 0; index != ne; index += 1) {
    s += sep;
    s += GenerationSettings::idl_type(type(etable->type(index), list, nodes));
    sep = ", ";
  }
  
  if (index != 0)
    s += ')';
}

void OperationDialog::idl_default_def() {
  QString s = oper->default_idl_decl(edname->text().stripWhiteSpace());
  QString params;
  int index;
  int nparams = (int) table->nparams();
  const char * sep;
  
  for (index = 0, sep = ""; index != nparams; index += 1, sep = ", ") {
    QString p;
    
    p.sprintf("%s${d%d} ${t%d} ${p%d}", sep, index, index, index);
    params += p;
  }
  
  if ((index = s.find("${)}")) != -1)
    s.insert(index, params);
  
  edidldecl->setText(s);
  idl_update_decl();
}

void OperationDialog::idl_unmapped_def() {
  edidldecl->setText(QString::null);
  showidldecl->setText(QString::null);
}

void OperationDialog::idl_update_decl() {  
  // do NOT write
  //	const char * p = edidldecl->text();
  // because the QString is immediatly destroyed !
  QString def = edidldecl->text();
  const char * p = def;
  QString indent = "";
  QString s;
  unsigned rank;
  
  while ((*p == ' ') || (*p == '\t'))
    indent += *p++;
  
  s = indent;
  
  while (*p) {
    if (!strncmp(p, "${comment}", 10))
      manage_comment(comment->text(), p, s, indent);
    else if (!strncmp(p, "${oneway}", 9)) {
      p += 9;
      if (oneway_cb->isChecked())
	s += "oneway ";
    }
    else if (!strncmp(p, "${type}", 7)) {
      p += 7;
      s += GenerationSettings::idl_type(type(edreturn_type->currentText().stripWhiteSpace(),
					     list, nodes));
    }
    else if (!strncmp(p, "${name}", 7)) {
      p += 7;
      s += compute_name(edidlnamespec);
    }
    else if (!strncmp(p, "${(}", 4)) {
      p += 4;
      s += '(';
    }
    else if (!strncmp(p, "${)}", 4)) {
      p += 4;
      s += ')';
    }
    else if (sscanf(p, "${d%u}", &rank) == 1) {
      manage_dir(rank, s);
      p = strchr(p, '}') + 1;
    }
    else if (sscanf(p, "${t%u}", &rank) == 1) {
      manage_idl_type(rank, s);
      p = strchr(p, '}') + 1;
    }
    else if (sscanf(p, "${p%u}", &rank) == 1) {
      manage_var(rank, s);
      p = strchr(p, '}') + 1;
    }
    else if (oper->is_get_or_set && !strncmp(p, "${stereotype}", 13)) {
      p += 13;
      // get/set relation with multiplicity > 1
      s += GenerationSettings::idl_relation_stereotype(oper->stereotype);
    }
    else if (!strncmp(p, "${raisesnl}", 11)) {
      if (etable->nexceptions() != 0) {
	s += '\n';
	s += indent;
	s += "  ";
      }
      p += 11;
    }
    else if (!strncmp(p, "${raises}", 9)) {
      manage_idl_exceptions(s);
      p += 9;
    }
    else if (*p == '\n') {
      s += *p++;
      if (*p)
	s += indent;
    }
    else
      s += *p++;
  }
  
  showidldecl->setText(s);
}

QString OperationDialog::idl_decl(const BrowserOperation * op)
{
  OperationData * d = (OperationData *) op->get_data();
  QCString decl = d->idl_decl;
  
  remove_comments(decl);
  
  const char * p = decl;
  QString s;
  unsigned rank;

  while ((*p == ' ') || (*p == '\t'))
    p += 1;
  
  while (*p) {
    if (!strncmp(p, "${comment}", 10))
      p += 10;
    else if (!strncmp(p, "${oneway}", 9))
      p += 9;
    else if (!strncmp(p, "${type}", 7)) {
      p += 7;
      s += GenerationSettings::idl_type(d->return_type.get_type());
    }
    else if (!strncmp(p, "${name}", 7)) {
      p += 7;
      s += op->compute_name(d->idl_name_spec);
    }
    else if (!strncmp(p, "${(}", 4)) {
      p += 4;
      s += '(';
    }
    else if (!strncmp(p, "${)}", 4)) {
      p += 4;
      s += ')';
    }
    else if (sscanf(p, "${d%u}", &rank) == 1) {
      if (rank < d->nparams) 
	switch (d->params[rank].get_dir()) {
	case UmlIn:
	  s += "in";
	  break;
	case UmlOut:
	  s += "out";
	  break;
	default:
	  s += "inout";
      }
      else {
	s += "${d";
	s += QString::number(rank);
	s += '}';
      }      
      p = strchr(p, '}') + 1;
    }
    else if (sscanf(p, "${t%u}", &rank) == 1) {
      if (rank < d->nparams) 
	s += GenerationSettings::idl_type(d->params[rank].get_type().get_type());
      else {
	s += "${t";
	s += QString::number(rank);
	s += '}';
      }      
      p = strchr(p, '}') + 1;
    }
    else if (sscanf(p, "${p%u}", &rank) == 1) {
      if (rank < d->nparams) 
	s += d->params[rank].get_name();
      else {
	s += "${p";
	s += QString::number(rank);
	s += '}';
      }
      p = strchr(p, '}') + 1;
    }
    else if (d->is_get_or_set && !strncmp(p, "${stereotype}", 13)) {
      p += 13;
      // get/set relation with multiplicity > 1
      s += GenerationSettings::idl_relation_stereotype(d->stereotype);
    }
    else if (!strncmp(p, "${raisesnl}", 11)) {
      p += 11;
      s += ' ';
    }
    else if (!strncmp(p, "${raises}", 9))
      p += 9;
    else if (*p == '\n') {
      s += ' ';
      do
	p+= 1;
      while ((*p == ' ') || (*p == '\t'));
    }
    else if (*p == ';')
      break;
    else
      s += *p++;
  }
  
  return s;
}

//
// ParamTable
//

// copy/cut/paste
QString ParamsTable::dir_copy;
QString ParamsTable::name_copy;
QString ParamsTable::type_copy;
QString ParamsTable::default_value_copy;

static QStringList DirList;

ParamsTable::ParamsTable(OperationData * o, QWidget * parent,
			 const QStringList & list, bool visit)
    : MyTable(o->get_n_params() + 1, (visit) ? 4 : 5, parent), types(list) {
  int index;
  int sup = o->get_n_params();
    
  setSorting(FALSE);
  setSelectionMode(NoSelection);	// single does not work
  setRowMovingEnabled(TRUE);
  horizontalHeader()->setLabel(0, "Direction");
  horizontalHeader()->setLabel(1, "Name");
  horizontalHeader()->setLabel(2, "Type");
  horizontalHeader()->setLabel(3, "Default value");
  
  if (visit) {
    for (index = 0; index < sup; index += 1){
      setItem(index, 0, new TableItem(this, QTableItem::Never, stringify(o->get_param_dir(index))));
      setItem(index, 1, new TableItem(this, QTableItem::Never, o->get_param_name(index)));
      setItem(index, 2, new TableItem(this, QTableItem::Never, o->get_param_type(index).get_full_type()));
      setItem(index, 3, new TableItem(this, QTableItem::Never, o->get_param_default_value(index)));
    }
  }
  else {
    horizontalHeader()->setLabel(4, "*");
    
    alltypes = GenerationSettings::basic_types();
    for (QStringList::ConstIterator it = list.begin(); it != list.end(); ++it)
      alltypes.append(*it);
    
    if (DirList.count() == 0) {
      // Made DirList without be linked with the items order
      DirList.append(QString::null);
      DirList.append(QString::null);
      DirList.append(QString::null);
      DirList[UmlInOut] = stringify(UmlInOut);
      DirList[UmlIn] = stringify(UmlIn);
      DirList[UmlOut] = stringify(UmlOut);
    }
    
    for (index = 0; index < sup; index += 1){
      setItem(index, 0, new ComboItem(this, stringify(o->get_param_dir(index)), DirList, FALSE));
      setText(index, 1, o->get_param_name(index));
      setItem(index, 2, new ComboItem(this, o->get_param_type(index).get_full_type(), alltypes));
      setText(index, 3, o->get_param_default_value(index));
      setText(index, 4, QString::null);
    }
    
    setItem(index, 0, new ComboItem(this, stringify(UmlInOut), DirList, FALSE));
    setText(index, 1, QString::null);
    setItem(index, 2, new ComboItem(this, QString::null, alltypes));
    setText(index, 3, QString::null);
    setText(index, 4, QString::null);
  }
  
  adjustColumn(0);
  //setColumnStretchable (1, TRUE);
  horizontalHeader()->setResizeEnabled(TRUE, 1);
  setColumnStretchable (2, TRUE);
  //setColumnStretchable (3, TRUE);
  horizontalHeader()->setResizeEnabled(TRUE, 3);
  
  if (! visit) {
    adjustColumn(4);
    setColumnStretchable (4, FALSE);
    
    connect(this, SIGNAL(pressed(int, int, int, const QPoint &)),
	    this, SLOT(button_pressed(int, int, int, const QPoint &)));
    connect(this, SIGNAL(valueChanged(int, int)),
	    this, SLOT(value_changed(int, int)));
  }
}

void ParamsTable::activateNextCell() {
  int row = currentRow();
  int col = currentColumn();
    
  if (col == 3) {
    // goto next line
    if (++row == numRows()) {
      // adds a new line
      setNumRows(row + 1);
      setItem(row, 0, new ComboItem(this, stringify(UmlInOut), DirList, FALSE));
      setText(row, 1, QString::null);
      setItem(row, 2, new ComboItem(this, QString::null, alltypes));
      setText(row, 3, QString::null);
      setText(row, 4, QString::null);
    }
    setCurrentCell(row, 0);
  }
  else
    setCurrentCell(row, col + 1);
}

void ParamsTable::value_changed(int row, int col) {
  if ((row == (numRows() - 1)) && (col != 0) && !text(row, col).isEmpty())
    insert_row_after(row);
}

void ParamsTable::button_pressed(int row, int col, int, const QPoint &) {
  if (col == 4) {
    char s[16];
    
    sprintf(s, "param %d", row + 1);
    
    QPopupMenu m;
    m.insertItem(s, -1);
    m.insertSeparator();
    m.insertItem("insert param before", 0);
    m.insertItem("insert param after", 1);
    m.insertSeparator();
    m.insertItem("delete param", 2);
    m.insertSeparator();
    m.insertItem("copy param", 3);
    m.insertItem("cut param", 4);
    m.insertItem("paste param", 5);
    m.insertSeparator();

    QPopupMenu mv;
    int rank;
    
    for (rank = 0; rank != numRows(); rank += 1)
      if (rank != row)
	mv.insertItem(QString::number(rank + 1), 10 + rank);
    
    m.insertItem("move param", &mv);
    
    switch (rank = m.exec(QCursor::pos())) {
    case 0:
      insert_row_before(row);
      break;
    case 1:
      insert_row_after(row);
      break;
    case 2:
      delete_row(row);
      break;
    case 3:
      copy_row(row);
      break;
    case 4:
      cut_row(row);
      break;
    case 5:
      paste_row(row);
      break;
    default:
      if (rank >= 10)
	move_row(row, rank - 10);
      break;
    }
  }
}

void ParamsTable::insert_row_before(int row) {
  int n = numRows();
  int index;
  
  setNumRows(n + 1);
  
  for (index = n; index != row; index -= 1) {
    QTableItem * it;
    
    it = item(index - 1, 0);
    takeItem(it);
    setItem(index, 0, it);
    setText(index, 1, text(index - 1, 1));
    it = item(index - 1, 2);
    takeItem(it);
    setItem(index, 2, it);
    setText(index, 3, text(index - 1, 3));
  }
  
  setItem(row, 0, new ComboItem(this, stringify(UmlInOut), DirList, FALSE));
  setText(row, 1, QString::null);
  setItem(row, 2, new ComboItem(this, QString::null, alltypes));
  setText(row, 3, QString::null);
  setText(row, 4, QString::null);
 }

void ParamsTable::insert_row_after(int row) {
  int n = numRows();
  int index;
  
  setNumRows(n + 1);
  
  for (index = n; index > row + 1; index -= 1) {
    QTableItem * it;
    
    it = item(index - 1, 0);
    takeItem(it);
    setItem(index, 0, it);
    setText(index, 1, text(index - 1, 1));
    it = item(index - 1, 2);
    takeItem(it);
    setItem(index, 2, it);
    setText(index, 3, text(index - 1, 3));
  }
  
  setItem(row + 1, 0, new ComboItem(this, stringify(UmlInOut), DirList, FALSE));
  setText(row + 1, 1, QString::null);
  setItem(row + 1, 2, new ComboItem(this, QString::null, alltypes));
  setText(row + 1, 3, QString::null);
  setText(row + 1, 4, QString::null);
 }

void ParamsTable::delete_row(int row) {
  int n = numRows();
  int index;

  clearCellWidget(row, 1);
    
  if (row == (n - 1)) {
    // the last line : empty it
    setItem(row, 0, new ComboItem(this, stringify(UmlInOut), DirList, FALSE));
    setText(row, 1, QString::null);
    setItem(row, 2, new ComboItem(this, QString::null, alltypes));
    setText(row, 3, QString::null);
  }
  else {
    for (index = row; index != n - 1; index += 1) {
      QTableItem * it;
      
      it = item(index + 1, 0);
      takeItem(it);
      setItem(index, 0, it);
      setText(index, 1, text(index + 1, 1));
      it = item(index + 1, 2);
      takeItem(it);
      setItem(index, 2, it);
      setText(index, 3, text(index + 1, 3));
    }
    setNumRows(n - 1);
  }
}

void ParamsTable::copy_row(int row) {
  dir_copy = text(row, 0);
  name_copy = text(row, 1);
  type_copy = text(row, 2);
  default_value_copy = text(row, 3);
}

void ParamsTable::cut_row(int row) {
  copy_row(row);
  delete_row(row);
}

void ParamsTable::paste_row(int row) {
  setText(row, 0, dir_copy);
  setText(row, 1, name_copy);
  setText(row, 2, type_copy);
  setText(row, 3, default_value_copy);
  
  if ((row == (numRows() - 1)) &&
      (!dir_copy.isEmpty() ||
       !name_copy.isEmpty() ||
       !type_copy.isEmpty()))
    insert_row_after(row);
}

void ParamsTable::move_row(int from, int to) {
  QString save_dir_copy = dir_copy;
  QString save_name_copy = name_copy;
  QString save_type_copy = type_copy;
  QString save_default_value_copy = default_value_copy;
  
  cut_row(from);
  if (to > from)
    insert_row_after(to - 1);
  else
    insert_row_before(to);
  paste_row(to);
  
  dir_copy = save_dir_copy;
  name_copy = save_name_copy;
  type_copy = save_type_copy;
  default_value_copy = save_default_value_copy;
}

void ParamsTable::update(OperationData * oper,
			 BrowserNodeList & nodes) {
  forceUpdateCells();
  
  int n = numRows();
  int index;
  
  if (text(n - 1, 1).isEmpty())
    n -= 1;
  
  oper->set_n_params(n);
  
  for (index = 0; index != n; index += 1) {    
    oper->set_param_dir(index, (UmlParamDirection) DirList.findIndex(text(index, 0)));
			
    oper->set_param_name(index, text(index, 1).stripWhiteSpace());
    
    AType t;
    QString s = text(index, 2).stripWhiteSpace();
    
    if (! s.isEmpty()) {
      int rank = types.findIndex(s);
      
      if (rank != -1)
	t.type = (BrowserClass *) nodes.at(rank);
      else
	t.explicit_type = s;
    }
     
    oper->set_param_type(index, t);
    
    oper->set_param_default_value(index, text(index, 3).stripWhiteSpace());
  }
}

unsigned ParamsTable::nparams() const {
  int n = numRows();
  
  return (text(n - 1, 1).isEmpty()) ? n -= 1 : n;
}

QString ParamsTable::name(unsigned rank) const {
  return text(rank, 1);
}

QString ParamsTable::type(unsigned rank) const {
  return text(rank, 2);
}

UmlParamDirection ParamsTable::dir(unsigned rank) const {
  return (UmlParamDirection) DirList.findIndex(text(rank, 0));
}

QString ParamsTable::value(unsigned rank) const {
  return text(rank, 3);
}

//
// ExceptionTable
//

// copy/cut/paste
QString ExceptionsTable::type_copy;

ExceptionsTable::ExceptionsTable(OperationData * o, QWidget * parent,
				 const QStringList & list, bool visit)
    : MyTable(o->get_n_exceptions() + 1, (visit) ? 1 : 2, parent), types(list) {
  int index;
  int sup = o->get_n_exceptions();
  
  setSorting(FALSE);
  setSelectionMode(NoSelection);	// single does not work
  setRowMovingEnabled(TRUE);
  horizontalHeader()->setLabel(0, "Type");
  
  if (visit) {
    for (index = 0; index < sup; index += 1)
      setItem(index, 0, new TableItem(this, QTableItem::Never, o->get_exception(index).get_full_type()));
  }
  else {
    horizontalHeader()->setLabel(1, "*");
    
    for (index = 0; index < sup; index += 1){
      setItem(index, 0, new ComboItem(this, o->get_exception(index).get_full_type(), types));
      setText(index, 1, QString::null);
    }
    
    setItem(index, 0, new ComboItem(this, QString::null, types));
    setText(index, 1, QString::null);
  }
  
  setColumnStretchable (0, TRUE);
  
  if (! visit) {
    adjustColumn(1);
    setColumnStretchable (1, FALSE);
    
    connect(this, SIGNAL(pressed(int, int, int, const QPoint &)),
	    this, SLOT(button_pressed(int, int, int, const QPoint &)));
    connect(this, SIGNAL(valueChanged(int, int)),
	    this, SLOT(value_changed(int, int)));
  }
}

void ExceptionsTable::activateNextCell() {
  int row = currentRow();
    
  // goto next line
  if (++row == numRows()) {
    // adds a new line
    setNumRows(row + 1);
    setItem(row, 0, new ComboItem(this, QString::null, types));
    setText(row, 1, QString::null);
  }
  setCurrentCell(row, 0);
}

void ExceptionsTable::value_changed(int row, int col) {
  if ((row == (numRows() - 1)) && !text(row, col).isEmpty())
    insert_row_after(row);
}

void ExceptionsTable::button_pressed(int row, int col, int, const QPoint &) {
  if (col == 1) {
    char s[16];
    
    sprintf(s, "exception %d", row + 1);
    
    QPopupMenu m;
    m.insertItem(s, -1);
    m.insertSeparator();
    m.insertItem("insert exception before", 0);
    m.insertItem("insert exception after", 1);
    m.insertSeparator();
    m.insertItem("delete exception", 2);
    m.insertSeparator();
    m.insertItem("copy exception", 3);
    m.insertItem("cut exception", 4);
    m.insertItem("paste exception", 5);
    m.insertSeparator();

    QPopupMenu mv;
    int rank;
    
    for (rank = 0; rank != numRows(); rank += 1)
      if (rank != row)
	mv.insertItem(QString::number(rank + 1), 10 + rank);
    
    m.insertItem("move exception", &mv);
    
    switch (rank = m.exec(QCursor::pos())) {
    case 0:
      insert_row_before(row);
      break;
    case 1:
      insert_row_after(row);
      break;
    case 2:
      delete_row(row);
      break;
    case 3:
      copy_row(row);
      break;
    case 4:
      cut_row(row);
      break;
    case 5:
      paste_row(row);
      break;
    default:
      if (rank >= 10)
	move_row(row, rank - 10);
      break;
    }
  }
}

void ExceptionsTable::insert_row_before(int row) {
  int n = numRows();
  int index;
  
  setNumRows(n + 1);
  
  for (index = n; index != row; index -= 1) {
    QTableItem * it = item(index - 1, 0);
    
    takeItem(it);
    setItem(index, 0, it);
  }
  
  setItem(row, 0, new ComboItem(this, QString::null, types));
  setText(row, 1, QString::null);
 }

void ExceptionsTable::insert_row_after(int row) {
  int n = numRows();
  int index;
  
  setNumRows(n + 1);
  
  for (index = n; index > row + 1; index -= 1) {
    QTableItem * it = item(index - 1, 0);
    
    takeItem(it);
    setItem(index, 0, it);
  }
  
  setItem(row + 1, 0, new ComboItem(this, QString::null, types));
  setText(row + 1, 1, QString::null);
 }

void ExceptionsTable::delete_row(int row) {
  int n = numRows();
  int index;

  clearCellWidget(row, 1);
    
  if (row == (n - 1)) {
    // the last line : empty it
    setItem(0, 0, new ComboItem(this, QString::null, types));
    setText(0, 1, QString::null);
  }
  else {
    for (index = row; index != n - 1; index += 1) {
      QTableItem * it = item(index + 1, 0);
      takeItem(it);
      setItem(index, 0, it);
    }
    setNumRows(n - 1);
  }
}

void ExceptionsTable::copy_row(int row) {
  type_copy = text(row, 0);
}

void ExceptionsTable::cut_row(int row) {
  copy_row(row);
  delete_row(row);
}

void ExceptionsTable::paste_row(int row) {
  setText(row, 0, type_copy);
  
  if ((row == (numRows() - 1)) && !type_copy.isEmpty())
    insert_row_after(row);
}

void ExceptionsTable::move_row(int from, int to) {
  QString save_type_copy = type_copy;
  
  cut_row(from);
  if (to > from)
    insert_row_after(to - 1);
  else
    insert_row_before(to);
  paste_row(to);
  
  type_copy = save_type_copy;
}

void ExceptionsTable::update(OperationData * oper,
			     BrowserNodeList & nodes) {
  forceUpdateCells();
  
  int n = numRows();
  int index;
  
  if (text(n - 1, 0).isEmpty())
    n -= 1;
  
  oper->set_n_exceptions(n);
  
  for (index = 0; index != n; index += 1) {   
    AType t;
    QString s = text(index, 0).stripWhiteSpace();
    
    if (!s.isEmpty()) {
      int rank = types.findIndex(s);
    
      if (rank != -1) 
	t.type = (BrowserClass *) nodes.at(rank);
      else
	t.explicit_type = s;
    }
    oper->set_exception(index, t);
  }
}

unsigned ExceptionsTable::nexceptions() const {
  int n = numRows();
  
  return (text(n - 1, 0).isEmpty()) ? n -= 1 : n;
}

QString ExceptionsTable::type(unsigned rank) const {
  return text(rank, 0);
}

